'use strict';

const EventEmitter = require('events');
const amqplib      = require('amqplib');
const isFatalError = require('amqplib/lib/connection').isFatalError;
const Promise      = require('bluebird');
const domain       = require('domain');

const RECONNET_INTERVAL_DEFAULT = 1; // in seconds

//TODO
// 1. queue published messages in case of failure in the connection
// 2. re subscribe in case of reconnection

class EventBus extends EventEmitter {

    constructor() {
        super();
        this.initialized = false;
    }

    _ensureEventbusInitializaed() {
        if (!this.initialized) {
            throw new Error('Eventbus was not initialized');
        }
    }

    _createErrorHandler() {
        if (this.errorHandlerCreated) {
            return;
        }

        this.errorHandlerCreated = true;
        this.on('error', (err) => {
            if (isFatalError(err)) { // reconnect only on fatal error
                this.reconnect = true;
                setTimeout(() => {
                    this.emit('reconnecting');
                    this.init(this.url, this.microServiceName, this.options);
                }, (this.options.reconnectInterval || RECONNET_INTERVAL_DEFAULT) * 1000);
            }
        });
    }

    quit() {
        this.connection.close().done();
    }

    init(url, microServiceName, options) {
        if (!microServiceName) {
            throw new Error('Micro service name missing');
        }

        this.url              = url;
        this.microServiceName = microServiceName;
        this.options          = options || {};

        this._createErrorHandler();

        this.openPromise = amqplib.connect(url);
        this.initialized = true;
        this.openPromise
            .then((newConnection) => {
                if (this.reconnect && this.connection) {
                    this.publishChannel = undefined;
                    this.emit('reconnected');
                }
                this.reconnect  = false;
                this.connection = newConnection;

                this.emit('ready');

                this.connection.on('error', (err) => {
                    this.emit('error', err);
                });

                this.connection.on('close', () => {
                    this.emit('close');
                });

                this.connection.on('blocked', () => {
                    this.emit('blocked');
                });

                this.connection.on('unblocked', () => {
                    this.emit('unblocked');
                });
            })
            .catch((err) => {
                this.emit('error', err);
            })
            .done();
    }

    publish(eventName, msg) {

        return Promise.resolve()
            .then(() => {
                this._ensureEventbusInitializaed();
                if (!this.publishChannel) {
                    return this._createPublishChannel();
                }
            })
            .then(() => {
                return this.publishChannel.assertExchange(eventName, 'fanout', { durable: true });
            })
            .then(() => {
                return this.publishChannel.publish(eventName, 'my-routing-key', new Buffer(this._serialize(msg)));
            });
    }

    subscribe(eventName, handler) {
        this._ensureEventbusInitializaed();
        const subscriberEmitter = new EventEmitter();

        const queueName = `${this.microServiceName}_${eventName}`;
        Promise.resolve()
            .then(() => {
                return this._getSubscribeChannel();
            })
            .then((ch) => {
                return ch.assertExchange(eventName, 'fanout', { durable: true })
                    .then(() => {
                        return ch.assertQueue(queueName, { durable: true })
                            .then(() => {
                                return ch.bindQueue(queueName, eventName);
                            })
                            .then(() => {
                                return ch.consume(queueName, (msg) => {
                                    const d = domain.create();
                                    d.on('error', (err) => {
                                        subscriberEmitter.emit('error', { handler: err });
                                    });
                                    d.run(() => {
                                        ch.ack(msg); // TODO allow controlling when to ack
                                        handler(this._deserialize(msg.content.toString()))
                                            .catch((err) => {
                                                subscriberEmitter.emit('error', { handler: err });
                                            })
                                            .done();
                                    });
                                });
                            });
                    });
            })
            .catch((err) => {
                subscriberEmitter.emit('error', { connection: err });
            })
            .done();

        return subscriberEmitter;
    }

    _createPublishChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            })
            .then((ch) => {
                this.publishChannel = ch;
            });
    }

    _getSubscribeChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            });
    }

    _serialize(body) {
        return JSON.stringify(body);
    }

    _deserialize(content) {
        return JSON.parse(content);
    }

}

module.exports = new EventBus();
